# -*- coding: utf8 -*-
"""
基于 concurrent.futures.thread.ThreadPoolExecutor 实现一个异步任务框架，对于长时间的任务直接交结异步构架执行，以便尽快的返回 JsonResponse 。

这里的实现参考了 JavaScript 的并发模型 "https://developer.mozilla.org/zh-CN/docs/Web/JavaScript/EventLoop"

"""


import logging
from uuid import uuid4
from concurrent.futures.thread import ThreadPoolExecutor

from dbmcenter.settings import BACKEND_THREAD_POOL_SIZE

threads = None

logger = logging.getLogger('dbmcenter.core.task')


class DbmBaseTask(object):
    """
    所有 dbm-center task 的基类
    """

    def __init__(self):
        """
        """
        self.task_id = uuid4()

    def check(self):
        """
        检查任务能否执行
        """
        raise NotImplementedError("DbmBaseTask.check NotImplemented")

    def save(self):
        """
        如果任务在 MySQL 中不存在，就接入
        如果任务在 MySQL 中已存在，就更新
        """
        raise NotImplementedError("DbmBaseTask.save NotImplemented")
    
    def __call__(self):
        """
        每个任务都是一个可以执行的对象
        """
        raise NotImplementedError("DbmBaseTask.__call__ NotImplemented")



def execute_task(task):
    """
    """
    logging.info(f"backend threads pool size = {BACKEND_THREAD_POOL_SIZE} .")
    try:
        task()
    except Exception as err:
        logger.error("execute task fail.")
        logger.error(str(err))


def submit_backend_task(task:DbmBaseTask):
    """
    添加后台执行任务
    """
    threads.submit(execute_task, task)


def init_task_framework():
    """
    初始化后台执行框架
    """
    global threads
    if threads is None:
        threads = ThreadPoolExecutor(max_workers=BACKEND_THREAD_POOL_SIZE)



# -*- coding: utf8 -*-

"""
dbm-center 可能会被运行在多台机器的多个进程上，虽然大家都可以处理 http 请求，但是对于关键状态应该只有一个进程可以读写。

"""

import time
import random
import logging

from dbmcenter.settings import IS_MASTER_ROLE_ENABLED
from dbmcenter.core.tasks.base import DbmBaseTask, threads, init_task_framework, submit_backend_task


logger = logging.getLogger('dbmcenter.core.tasks')


class RegisterMasterTask(DbmBaseTask):
    """
    注册 dbm-master 结点信息
    """

    def register_master(self):
        """
        把自己注册为 master
        """
        # 对于 swgi 这种多进程的场景，这里随机一下不要让它们同时去抢主
        time.sleep(random.randint(3, 29))

        # 抢到主的进程不断的序约，没有抢到的进程也不会放
        # 序约 13 秒一次
        # 抢主 27 秒一次
        while True:
            logger.info("register-master:")
            # TODO
            time.sleep(13)

    def __call__(self):
        self.register_master()


def init_master():
    """
    初始化 master 角色
    """
    # 如果不启动 MASTER 状态就直接退出
    if not IS_MASTER_ROLE_ENABLED:
        return
    
    global threads
    if threads is None:
        # 激活异步任务框架
        init_task_framework()
        submit_backend_task(RegisterMasterTask())



# -*- coding: utf8 -*-

"""
返回 dbm-center 的配置信息方便查询
"""
import os
import logging
from django.utils import timezone
from django.http import JsonResponse
from django.views.generic.base import View
from dbmcenter import settings

logger = logging.getLogger('dbmcenter.views.configs')

class DBMCenterConfigView(View):
    """返回 DBMCenter 的配置信息
    """
    def get(self, request,  *args, **kwargs):
        """读取 settings.py 中的部分配置返回给客户端
        """
        logger.info(f"path = {request.path} method= {request.method} args = {args}  kwargs = {kwargs} .")

        # 读取配置
        basedir = str(settings.BASE_DIR)
        engine = settings.DATABASES['default']['ENGINE']
        name = ''
        if engine == 'django.db.backends.sqlite3':
            name = os.path.join(basedir, 'dbmcenter.sqlite3')
        
        # 返回结果
        return JsonResponse(data={
            'BASE_DIR': basedir,
            'ENGINE': engine,
            'NAME': name,
            'CURRENT': timezone.now()
        })


from django.db import models
from django.utils import timezone

import logging
# Create your models here.

logger = logging.getLogger("agents")

class Agent(models.Model):
    """
    用于保留系统里面每一个 Agent 的信息
    """

    # 心跳上报的超时时间
    HEARTBEAT_EXPIRE_TIME_SECONDES = 11

    # agent 所在主机的地址/域名
    host = models.GenericIPAddressField(unique=True, default='127.0.0.1')

    # agent 的版本号
    version = models.CharField(max_length=16)

    # 端口
    port = models.IntegerField(default=8086)

    # 注册到系统的时间
    register_at = models.DateTimeField(default=timezone.now)

    # 最近一次上报心跳的时间点
    heartbeat_at = models.DateTimeField(default=timezone.now)

    @property
    def is_alive(self):
        """
        heartbeat_at 用于保存 agent 上报心跳时的时间点，is_alive 去检查这个时间点，如果这个时间点距离当前时间点小于 11s 就算活着的。
        """
        now = timezone.now()
        logger.debug(f"check agent is alive , now = {now} agent.heartbeat_at = {self.heartbeat_at} .")
        if (now - self.heartbeat_at) > timezone.timedelta(seconds=self.HEARTBEAT_EXPIRE_TIME_SECONDES):
            logger.warning(f"agent is not alive (now - self.heartbeat_at) = {now - self.heartbeat_at}. ")
            return False

        return True 

    @property
    def is_connectble(self):
        """
        dbm-center 能否连接到 dbm-agent,
        """
        return False

    def update_heartbeat(self):
        """
        更新 agent 的心跳信息
        """
        self.heartbeat_at = timezone.now()
        self.save()



from django.test import TestCase, Client
from django.utils import timezone


from .models import Agent
# Create your tests here.


class AgentsViewTestCase(TestCase):
    """
    """
    
    def setup(self):
        """
        """
        pass

    def test_given_agents_is_empty_when_query_all_then_return_empty(self):
        """
        given: 数据库中没有任何一条 agent 的记录
        when: 查询所有 agent 信息的时候
        then: 返回一个空的列表

        期待的返回如下
        {
            agent:[],
            message:''
        }
        """
        client = Client()
        response = client.get("/apis/agents/")
        self.assertEqual(response.status_code, 200)

        data = response.json()
        self.assertEqual(len(data['agents']), 0)

        self.assertIn('message', data)

class AgentModelTestCase(TestCase):
    """
    针对 Agent 这个 model 的测试
    """

    def setUp(self) -> None:
        """
        """
        now = timezone.now()
        self.agent_not_time_out = Agent.objects.create(host="127.0.0.1", heartbeat_at=now - timezone.timedelta(seconds=10))
        self.agent_time_out = Agent.objects.create(host="127.0.0.2", heartbeat_at=now - timezone.timedelta(seconds=12))

    def test_given_agent_heartbeat_at_when_check_agent_is_alive_then_return_false(self):
        """
        given: agent 心跳上报超时(数据库里面保存的值与，当前时间之前差大于 11s )
        when: 检查 agent 是不是活着的
        then: Flase
        """
        # now = timezone.now()
        # now = now - timezone.timedelta(seconds=12)
        # agent = Agent.(heartbeat_at=now)
        self.assertFalse(self.agent_time_out.is_alive)

    def test_given_agent_heartheat_at_not_timeout_when_check_agent_is_alive_then_return_true(self):
        """
        given: agent 的心路上报时间为 10 秒以前
        when: 检查 agent 是不是活着的
        then: True
        """
        now = timezone.now()
        now = now - timezone.timedelta(seconds=10)
        agent = Agent(heartbeat_at=now)
        self.assertTrue(self.agent_not_time_out.is_alive)



# -*- coding: utf8 -*-

import json
import logging

from django.utils import timezone
from django.http import JsonResponse
from django.db.utils import IntegrityError
from django.views.generic.base import View
from django.forms.models import model_to_dict

from agents.models import Agent
from dbmcenter.consts import APPLICATION_JSON

logger = logging.getLogger('agents')


class AgentsView(View):
    """
    所有 agent 对象 crud 的接口
    """

    def get(self, request, agent_pk='', *args, **kwargs):
        """
        根据 agent_pk 查询出对应的 agent 信息并返回，如果 agent_pk = '' 就返回所有的 agent 对象。

        Parameters
        ----------
        request : WSGIRequest
            请求对象

        agent_pk: str
            接口传进来的 agent 的主键，由于项目使用的 re 来匹配参数，所以在接口没有传入参数时会得到空值

        args: tuple
            位置参数、默认 ()

        kwargs: disct
            关键字参数、默认 {}


        Returns
        -------
        result : JsonReponse
            返回所有 agent 的列表
        """
        logger.info(
            f"path = {request.path} method= {request.method} agent-pk = '{agent_pk}' args = {args}  kwargs = {kwargs} .")
        result = []
        message = ''

        # 处理有传入主键的情况
        if agent_pk != '':
            logger.info(f"query one agent by pk = {agent_pk}.")
            pk = int(agent_pk)

            try:
                agent = Agent.objects.get(pk=pk)
                return JsonResponse(model_to_dict(agent))
            except Agent.DoesNotExist as err:
                # Agent matching query does not exist.
                logger.warning(str(err))
                return JsonResponse({
                    'message': str(err)
                })

        # 处理查询所有的情况
        try:
            logger.info(f"query all agents .")
            result = [_ for _ in Agent.objects.values()]
            return JsonResponse({
                'agents': result,
                'message': message
            })

        except Exception as err:
            message = f"query all agents got fail, inner error = {str(err)}"
            logger.error(message)

    def post(self, request, *args, **kwargs):
        """
        agent 首次注册的时候会调用这个 post 方法

        Parameters
        ----------
        request : WSGIRequest
            请求对象

        args: tuple
            位置参数、默认 ()

        kwargs: disct
            关键字参数、默认 {}
        """
        # 检查 header
        logger.info(
            f"path = {request.path} method= {request.method} args = {args}  kwargs = {kwargs} .")
        if request.headers["Content-Type"] != APPLICATION_JSON:
            logger.warning(
                f"request.headers['Content-Type'] != '{APPLICATION_JSON}', we well reject it .")
            return JsonResponse({
                'message': 'pelease use application/json post your data.'
            })

        # 提取数据
        logger.info(f"start translate playload to json .")
        data = json.loads(request.body.decode('utf8'))
        logger.info(f"data = {data} .")

        # 业务逻辑的检查
        host = data.get('host', '')
        port = data.get('port', 8086)
        version = data.get('version', '0.0.0')
        now = timezone.now()
        if host == '':
            return JsonResponse({
                'message': 'host is empty , not valide .'
            })
        logger.info(f"host = {host} port = {port} version = {version} .")

        # 写回到数据库
        try:
            agent = Agent.objects.create(
                host=host, port=port, version=version, register_at=now, heartbeat_at=now)
        except IntegrityError as err:
            # 由于  host 要保证唯一，当提交上来的 host 已经存在的时候就会引发完整性约束异常
            logger.warning(f"looks like host = {host} , has been exists.")
            return JsonResponse({
                'message': str(err)
            })
        except Exception as err:
            logger.warning(type(err))
            return JsonResponse({
                'message': str(err)
            })

        # 返回 agent 的 id 值到 client
        return JsonResponse({
            'id': agent.pk,
            'message': '',
        })

    def put(self, request, agent_pk, *args, **kwargs): 
        """
        更新 agent 的心跳信息

        Parameter
        ---------
        agent_pk: int
            agent 的主键

        Return:
        ------
        {
            'pk': agent.pk,
            'message':''
        }
        """
        logger.info(f"{request} {agent_pk} {args} {kwargs}")

        if agent_pk != '':
            pk = int(agent_pk)
        else:
            return JsonResponse({
                'message': 'agent-pk should be int .'
            })

        # 更新 heartbeat
        try:
            agent = Agent.objects.get(pk=pk)
            agent.update_heartbeat()
        except Agent.DoesNotExist as err:
            return JsonResponse({
                'message': str(err)
            })
        # 
        return JsonResponse({
            'pk': agent.pk,
            'message': ''
            })



from setuptools import setup


setup(name='dbm-center',
      version='0.2.3',
      description='dbm-center 数据库管理中心服务端程序',
      author="Neeky",
      author_email="neeky@live.com",
      maintainer='Neeky',
      maintainer_email='neeky@live.com',
      scripts=['bin/dbm-center'],
      packages=['dbmcenter', 'agents', 'dbmcenter/core', 'dbmcenter/core/tasks', 'dbmcenter/views'],
      package_data={'dbmcenter': ['dbm-center/*', 'dbm-center/css/*', 'dbm-center/js/*']},
      install_requires=['Django>=4.1.2'],
      url='https://github.com/Neeky/dbm-center',
      python_requires='>=3.6.*',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Operating System :: POSIX',
          'Operating System :: MacOS :: MacOS X',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8']
      )


"""dbmcenter URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/4.1/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.urls import include, path
from django.http import HttpResponseRedirect

from dbmcenter.core.tasks.master import init_master
from dbmcenter.views.configs import DBMCenterConfigView

def index(request, *args, **kwargs):
    """
    重定向到 VUE 前端
    """
    return HttpResponseRedirect('dbm-center/index.html')


urlpatterns = [
    path('', index, name='index'),
    path('apis/agents/', include('agents.urls'), name='apis-agents'),
    path('apis/configs/', DBMCenterConfigView.as_view(), name='dbmcenter-configs')
]

init_master()